from typing import Optional

from indi.message.base import IndiMessage


class GetProperties(IndiMessage):
    from_device = True
    from_client = True

    def __init__(
        self,
        version: str,
        device: Optional[str] = None,
        name: Optional[str] = None,
        **junk
    ):
        super().__init__(device)
        self.version = version
        self.name = name



from typing import Optional

from indi.device import Driver, properties
from indi.device.events import Change, on
from indi.device.properties import standard
from indi.message import GetProperties, IndiMessage
from indi.transport.client.tcp import TCP as TCPClient


class Proxy(Driver):
    """Virtual proxy device driver

    Connects to another INDI server and passes all messages in both ways.
    All devices from remote server become visible in INDIpy.
    """

    address: Optional[str] = None
    port = 7624

    general = properties.Group(
        "GENERAL",
        vectors=dict(
            connection=standard.common.Connection(),
        ),
    )

    def __init__(self, *args, **kwargs):
        self._connection = None
        self._client = TCPClient(self.address, self.port)
        super().__init__(*args, **kwargs)

    def message_from_client(self, message: IndiMessage):
        if getattr(message, "device", None) in (None, self.name):
            super().message_from_client(message)

        if self._connection:
            self._connection.send_message(message)

    def _message_from_server(self, message: IndiMessage):
        if self._router:
            self._router.process_message(message, self)

    def accepts(self, device: Optional[str]):
        return True

    @on(general.connection.connect, Change)
    def connect(self, event):
        connected = self.get_group("general").connection.connect.bool_value
        if connected:
            self._connection = self._client.connect(self._message_from_server)
            self._connection.send_message(GetProperties(version="1.0"))
        else:
            if self._connection:
                self._connection.close()
            self._connection = None

    @property
    def client(self):
        return self._client



from typing import NewType

PermissionsType = NewType("PermissionsType", str)
StateType = NewType("StateType", str)
SwitchRuleType = NewType("SwitchRuleType", str)
SwitchStateType = NewType("SwitchStateType", str)
BLOBEnableType = NewType("BLOBEnableType", str)


class Permissions:
    READ_ONLY = PermissionsType("ro")
    WRITE_ONLY = PermissionsType("wo")
    READ_WRITE = PermissionsType("rw")


class State:
    IDLE = StateType("Idle")
    OK = StateType("Ok")
    BUSY = StateType("Busy")
    ALERT = StateType("Alert")


class SwitchRule:
    ONE_OF_MANY = SwitchRuleType("OneOfMany")
    AT_MOST_ONE = SwitchRuleType("AtMostOne")
    ANY_OF_MANY = SwitchRuleType("AnyOfMany")


class SwitchState:
    ON = SwitchStateType("On")
    OFF = SwitchStateType("Off")


class BLOBEnable:
    NEVER = BLOBEnableType("Never")
    ALSO = BLOBEnableType("Also")
    ONLY = BLOBEnableType("Only")



from indi.message.const import *


class DriverInterface:
    """Driver interface types
    Based on http://docs.indilib.org/drivers/driver-interface.html
    """

    GENERAL = 0
    TELESCOPE = 1 << 0
    CCD = 1 << 1
    GUIDER = 1 << 2
    FOCUSER = 1 << 3
    FILTER = 1 << 4
    DOME = 1 << 5
    GPS = 1 << 6
    WEATHER = 1 << 7
    AO = 1 << 8
    DUSTCAP = 1 << 9
    LIGHTBOX = 1 << 10
    DETECTOR = 1 << 11
    ROTATOR = 1 << 12
    SPECTROGRAPH = 1 << 13
    CORRELATOR = 1 << 14
    AUX = 1 << 15



from indi.device import properties
from indi.device.properties import const


def Exposure(
    label=None,
    default=0,
    min=0,
    max=2000,
    step=0.00000001,
    state=const.State.OK,
    perm=const.Permissions.READ_WRITE,
    timeout=0,
    enabled=True,
):
    return properties.NumberVector(
        name="CCD_EXPOSURE",
        elements=dict(
            time=properties.Number(
                name="CCD_EXPOSURE_VALUE", default=default, min=min, max=max, step=step
            )
        ),
    )


def UploadMode(
    label=None,
    default="UPLOAD_CLIENT",
    state=const.State.OK,
    perm=const.Permissions.READ_WRITE,
    timeout=0,
    enabled=True,
):
    return properties.SwitchVector(
        name="UPLOAD_MODE",
        default_on=default,
        rule=const.SwitchRule.ONE_OF_MANY,
        elements=dict(
            client=properties.Switch(name="UPLOAD_CLIENT"),
            local=properties.Switch(name="UPLOAD_LOCAL"),
            both=properties.Switch(name="UPLOAD_BOTH"),
        ),
    )



import sys
from functools import reduce

import indi
from indi.device import properties
from indi.device.properties import const


def Connection(
    label=None,
    default="DISCONNECT",
    state=const.State.OK,
    perm=const.Permissions.READ_WRITE,
    timeout=0,
    enabled=True,
):
    return properties.SwitchVector(
        name="CONNECTION",
        default_on=default,
        elements=dict(
            connect=properties.Switch(name="CONNECT"),
            disconnect=properties.Switch(name="DISCONNECT"),
        ),
    )


def DriverInfo(name=None, exec=None, version=None, interface=None):
    name = name or "INDIpy Driver"
    exec = exec or sys.argv[0]
    version = version or indi.__version__
    interface = interface or []
    interface = reduce(lambda a, b: a | b, interface, const.DriverInterface.GENERAL)

    elements = {
        "name": properties.TextVector.element_class(name="DRIVER_NAME", default=name),
        "exec": properties.TextVector.element_class(name="DRIVER_EXEC", default=exec),
        "version": properties.TextVector.element_class(
            name="DRIVER_VERSION", default=version
        ),
        "interface": properties.TextVector.element_class(
            name="DRIVER_INTERFACE", default=interface
        ),
    }

    return properties.TextVector(
        "DRIVER_INFO", elements=elements, perm=const.Permissions.READ_ONLY
    )



from typing import Optional

from indi.message import IndiMessage


class Device:
    def message_from_client(self, message: IndiMessage):
        raise Exception("Not implemented")

    def accepts(self, device: Optional[str]) -> bool:
        raise Exception("Not implemented")



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Tuple, Type, Union

from indi.device import events, values
from indi.device.events import EventSource
from indi.message import checks, const, def_parts, one_parts

if TYPE_CHECKING:
    from indi.device.driver import Device
    from indi.device.properties.instance.vectors import Vector


logger = logging.getLogger(__name__)


class Element(EventSource):
    def_message_class: Union[
        Type[def_parts.DefSwitch],
        Type[def_parts.DefNumber],
        Type[def_parts.DefLight],
        Type[def_parts.DefBLOB],
        Type[def_parts.DefText],
    ]
    set_message_class: Union[
        Type[one_parts.OneSwitch],
        Type[one_parts.OneNumber],
        Type[one_parts.OneLight],
        Type[one_parts.OneBLOB],
        Type[one_parts.OneText],
    ]
    allowed_value_types: Tuple[Type, ...] = (None.__class__,)

    def __init__(self, vector: Vector, definition):
        self._vector = vector
        self._definition = definition
        self._value = definition.default
        self._enabled = definition.enabled

    @property
    def name(self) -> str:
        return self._definition.name

    @property
    def device(self) -> Device:
        return self._vector.device

    @property
    def vector(self) -> Vector:
        return self._vector

    @property
    def enabled(self) -> bool:
        return self._enabled

    @enabled.setter
    def enabled(self, value: bool):
        self._enabled = value

    @property
    def value(self):
        e = events.Read(element=self)
        self.raise_event(e)
        return self._value

    @value.setter
    def value(self, value):
        logger.debug(
            "Element: setting value of element %s to %s", self._definition.name, value
        )
        prev_value = self._value
        self.check_value_type(value)
        self._value = self.check_value(value)
        self.device.send_message(self._vector.to_set_message())

        if prev_value != self._value:
            e = events.Change(element=self, old_value=prev_value, new_value=self._value)
            self.raise_event(e)

    def set_value(self, value):
        e = events.Write(element=self, new_value=value)
        self.raise_event(e)

        if not e.prevent_default:
            self.value = value

    def set_value_from_message(self, msg):
        self.set_value(msg.value)

    def reset_value(self, value):
        self._value = self.check_value(value)

    def check_value(self, value):
        return value

    def check_value_type(self, value):
        assert isinstance(
            value, self.allowed_value_types
        ), f"Value of {self.name} should be of type {self.allowed_value_types}"

    def to_def_message(self):
        return self.def_message_class(
            name=self._definition.name,
            value=self.value,
            label=self._definition.label,
        )

    def to_set_message(self):
        return self.set_message_class(name=self._definition.name, value=self.value)


class Number(Element):
    def_message_class = def_parts.DefNumber
    set_message_class = one_parts.OneNumber
    allowed_value_types = (
        int,
        float,
    ) + Element.allowed_value_types

    def to_def_message(self):
        return self.def_message_class(
            name=self._definition.name,
            value=values.num_to_str(self.value, self._definition.format),
            label=self._definition.label,
            format=self._definition.format,
            min=self._definition.min,
            max=self._definition.max,
            step=self._definition.step,
        )

    def to_set_message(self):
        return self.set_message_class(
            name=self._definition.name,
            value=values.num_to_str(self.value, self._definition.format),
        )

    def set_value_from_message(self, msg):
        self.set_value(values.str_to_num(msg.value, self._definition.format))


class Text(Element):
    def_message_class = def_parts.DefText
    set_message_class = one_parts.OneText
    allowed_value_types = (str,) + Element.allowed_value_types


class Switch(Element):
    def_message_class = def_parts.DefSwitch
    set_message_class = one_parts.OneSwitch
    allowed_value_types = (str,) + Element.allowed_value_types

    def check_value(self, value):
        value = checks.dictionary(value, const.SwitchState)
        return self._vector.apply_rule(self, value)

    @property
    def bool_value(self):
        return self.value == const.SwitchState.ON

    @bool_value.setter
    def bool_value(self, value):
        self.value = const.SwitchState.ON if value else const.SwitchState.OFF

    def reset_bool_value(self, value):
        self._value = const.SwitchState.ON if value else const.SwitchState.OFF


class Light(Element):
    def_message_class = def_parts.DefLight
    set_message_class = one_parts.OneLight
    allowed_value_types = (str,) + Element.allowed_value_types

    def check_value(self, value):
        return checks.dictionary(value, const.State)


class BLOB(Element):
    def_message_class = def_parts.DefBLOB
    set_message_class = one_parts.OneBLOB
    allowed_value_types = (values.BLOB,) + Element.allowed_value_types

    def to_set_message(self):
        if self.value is None:
            return self.set_message_class(
                name=self._definition.name, value=None, format=None, size=None
            )
        return self.set_message_class(
            name=self._definition.name,
            value=self.value.binary_base64,
            format=self.value.format,
            size=self.value.size,
        )

    def set_value_from_message(self, msg):
        blob_value = values.BLOB.from_base64(msg.value, msg.format)
        assert msg.size == blob_value.size
        self.set_value(blob_value)



from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Tuple, Type, Union

from indi.message import checks, const
from indi.message.base import IndiMessage
from indi.message.one_parts import (
    IndiMessagePart,
    OneBLOB,
    OneLight,
    OneNumber,
    OneSwitch,
    OneText,
)

if TYPE_CHECKING:
    from indi.message import TimestampType


class SetVector(IndiMessage):
    child_class: Union[
        Type[OneBLOB], Type[OneLight], Type[OneNumber], Type[OneSwitch], Type[OneText]
    ]
    from_device = True

    def __init__(
        self,
        device: str,
        name: str,
        state: const.StateType,
        timeout: Optional[float] = None,
        timestamp: Optional[TimestampType] = None,
        message: Optional[str] = None,
        children: Optional[Tuple[IndiMessagePart, ...]] = None,
        **junk,
    ):
        super().__init__(device)
        self.name = name
        self.state = checks.dictionary(state, const.State)
        self.timeout = timeout
        self.timestamp = timestamp
        self.message = message
        self.children = checks.children(children, self.child_class)


class SetBLOBVector(SetVector):
    child_class = OneBLOB


class SetLightVector(SetVector):
    child_class = OneLight


class SetNumberVector(SetVector):
    child_class = OneNumber


class SetSwitchVector(SetVector):
    child_class = OneSwitch


class SetTextVector(SetVector):
    child_class = OneText



from typing import Dict, List, Type

from indi.device import Driver
from indi.routing import Router


class DevicePool:
    """Device driver registry.

    In most cases you want to use preinitialized `default_pool`.
    """

    def __init__(self) -> None:
        self.device_classes: List[type] = []
        self.devices: Dict[str, Driver] = {}

    def register(self, device_class: Type[Driver]) -> Type[Driver]:
        """Register device

        Register device driver class in device pool.
        Designed to be used as decorator.

        :param device_class: Device to be registered
        :type device_class: Type[Driver]
        :return: Unmodified device class
        :rtype: Type[Driver]
        """
        self.device_classes.append(device_class)
        return device_class

    def init(self, router: Router):
        """Initialize all devices registered in the device pool

        :param router: Router instance responsible for handling communication
        :type router: Router
        """
        for device_class in self.device_classes:
            dev = device_class(router=router)
            self.devices[dev.name] = dev

    def get(self, name: str) -> Driver:
        """Get device driver instance by name

        :param name: Device name
        :type name: str
        :return: Device driver instance
        :rtype: Driver
        """
        return self.devices[name]


default_pool = DevicePool()



from typing import Any, Optional, Type, Union

from indi.device.events import EventSourceDefinition
from indi.device.properties.instance import elements as instance_elements
from indi.device.properties.instance.vectors import Vector
from indi.message import const


class Element(EventSourceDefinition):
    instance_class: Union[
        Type[instance_elements.BLOB],
        Type[instance_elements.Light],
        Type[instance_elements.Number],
        Type[instance_elements.Switch],
        Type[instance_elements.Text],
    ]
    default_value: Any

    def __init__(
        self,
        name: str,
        label: Optional[str] = None,
        default: Any = None,
        enabled: bool = True,
    ):
        super().__init__()
        self.name = name
        self.label = label or name
        self.default = self.default_value if default is None else default
        self.enabled = enabled

    def instance(self, vector: Vector):
        return self.instance_class(vector, self)


class Number(Element):
    instance_class = instance_elements.Number
    default_value = 0.0

    def __init__(
        self,
        *args,
        format="%f",
        min: Optional[float] = None,
        max: Optional[float] = None,
        step: float = 0,
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.format = format
        self.min = min
        self.max = max
        self.step = step


class Text(Element):
    instance_class = instance_elements.Text
    default_value = ""


class Switch(Element):
    instance_class = instance_elements.Switch
    default_value = const.SwitchState.OFF


class Light(Element):
    instance_class = instance_elements.Light
    default_value = const.State.OK


class BLOB(Element):
    instance_class = instance_elements.BLOB
    default_value = None



from typing import Optional

from indi.message import checks, const
from indi.message.base import IndiMessagePart


class DefIndiMessagePart(IndiMessagePart):
    def __init__(self, name: str, value=None, label: Optional[str] = None, **junk):
        super().__init__(name=name, value=value)
        self.label = label


class DefBLOB(DefIndiMessagePart):
    pass


class DefLight(DefIndiMessagePart):
    def check_value(self, value):
        return checks.dictionary(value, const.State)


class DefNumber(DefIndiMessagePart):
    def __init__(
        self,
        name: str,
        format,
        min: float,
        max: float,
        step: float,
        value=None,
        label: Optional[str] = None,
        **junk
    ):
        super().__init__(name=name, value=value, label=label)
        self.format = format
        self.min = min
        self.max = max
        self.step = step

    def check_value(self, value):
        return checks.number(value)


class DefSwitch(DefIndiMessagePart):
    def check_value(self, value):
        return checks.dictionary(value, const.SwitchState)


class DefText(DefIndiMessagePart):
    pass



import asyncio
import logging
from typing import Optional

import aiofiles
from aiofiles.threadpool.text import AsyncTextIndirectIOWrapper

from indi.message import IndiMessage
from indi.routing import Client, Router
from indi.transport import Buffer

logger = logging.getLogger(__name__)


class ConnectionHandler(Client):
    def __init__(
        self,
        router: Router,
        stdin: AsyncTextIndirectIOWrapper,
        stdout: AsyncTextIndirectIOWrapper,
    ) -> None:
        self.stdin = stdin
        self.stdout = stdout
        self.buffer = Buffer()
        self.router = router
        if self.router:
            self.router.register_client(self)

    async def handle(self):
        try:
            await self.wait_for_messages()
        except:
            logger.exception("Error in client handler loop")

        logger.info("Stopping INDIpy server on TTY")
        self.close()

    async def wait_for_messages(self):
        while True:
            logger.debug(f"Waiting for data")
            message = await self._read()
            if not message:
                logger.debug(f"No data, exiting")
                break
            logger.debug("Received data: %s", message)
            self.buffer.append(message)
            self.buffer.process(self.message_from_client)

    def message_from_client(self, message: IndiMessage):
        self.router.process_message(message, sender=self)

    def message_from_device(self, message: IndiMessage):
        data = message.to_string().decode("latin1")
        logger.debug("Sending data: %s", data)
        asyncio.get_running_loop().create_task(self._write(data))

    def close(self):
        self.router.unregister_client(self)

    async def _read(self) -> str:
        return await self.stdin.readline()

    async def _write(self, data: str):
        await self.stdout.write(data)
        await self.stdout.flush()


class TTY:
    def __init__(
        self,
        router: Router,
        stdin: Optional[AsyncTextIndirectIOWrapper] = None,
        stdout: Optional[AsyncTextIndirectIOWrapper] = None,
    ) -> None:
        self.router = router
        self.stdin: AsyncTextIndirectIOWrapper = stdin or aiofiles.stdin
        self.stdout: AsyncTextIndirectIOWrapper = stdout or aiofiles.stdout

    async def start(self):
        logger.info("Starting INDIpy server on TTY")
        conn_handler = ConnectionHandler(self.router, self.stdin, self.stdout)
        await conn_handler.handle()



from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Tuple, Type, Union

from indi.message import checks, const
from indi.message.base import IndiMessage
from indi.message.def_parts import (
    DefBLOB,
    DefIndiMessagePart,
    DefLight,
    DefNumber,
    DefSwitch,
    DefText,
)

if TYPE_CHECKING:
    from indi.message import TimestampType


class DefVector(IndiMessage):
    from_device = True
    children_class: Union[
        Type[DefBLOB], Type[DefLight], Type[DefNumber], Type[DefSwitch], Type[DefText]
    ]

    def __init__(
        self,
        device: str,
        name: str,
        state: const.StateType,
        label: Optional[str] = None,
        group: Optional[str] = None,
        timestamp: Optional[TimestampType] = None,
        message: Optional[str] = None,
        children: Optional[Tuple[DefIndiMessagePart, ...]] = None,
        **junk,
    ):
        super().__init__(device)
        self.name = name
        self.state = checks.dictionary(state, const.State)
        self.label = label
        self.group = group
        self.timestamp = timestamp
        self.message = message
        self.children = checks.children(children, self.children_class)


class DefWritableVector(DefVector):
    def __init__(
        self,
        *args,
        perm: const.PermissionsType,
        timeout: Optional[TimestampType] = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.perm = checks.dictionary(perm, const.Permissions)
        self.timeout = timeout


class DefBLOBVector(DefWritableVector):
    children_class = DefBLOB


class DefLightVector(DefVector):
    children_class = DefLight


class DefNumberVector(DefWritableVector):
    children_class = DefNumber


class DefSwitchVector(DefWritableVector):
    children_class = DefSwitch

    def __init__(self, *args, rule: const.SwitchRuleType, **kwargs):
        super().__init__(*args, **kwargs)
        self.rule = checks.dictionary(rule, const.SwitchRule)


class DefTextVector(DefWritableVector):
    children_class = DefText



from indi.message import checks, const
from indi.message.base import IndiMessagePart


class OneBLOB(IndiMessagePart):
    def __init__(self, name: str, size: float, format: str, value, **junk):
        super().__init__(name=name, value=value)
        self.size = size
        self.format = format


class OneLight(IndiMessagePart):
    def check_value(self, value):
        return checks.dictionary(value, const.State)


class OneNumber(IndiMessagePart):
    def check_value(self, value):
        return checks.number(value)


class OneSwitch(IndiMessagePart):
    def check_value(self, value):
        return checks.dictionary(value, const.SwitchState)


class OneText(IndiMessagePart):
    pass



from typing import Optional

from indi.message import checks, const
from indi.message.base import IndiMessage


class EnableBLOB(IndiMessage):
    from_client = True

    def __init__(
        self,
        device: str,
        value: const.BLOBEnableType,
        name: Optional[str] = None,
        **junk
    ):
        super().__init__(device)
        self.name = name
        self.value = checks.dictionary(value, const.BLOBEnable)



from .definition.elements import BLOB, Light, Number, Switch, Text
from .definition.group import Group
from .definition.vectors import (
    BLOBVector,
    LightVector,
    NumberVector,
    SwitchVector,
    TextVector,
)



from indi.message.base import IndiMessage


class OneLight(IndiMessage):
    from_device = True

    def __init__(self, name: str, value, **junk):
        super().__init__()
        self.name = name
        self.value = value



import logging
from typing import Dict, List, Optional, Union, cast

from indi.message import EnableBLOB, IndiMessage, NewBLOBVector, const
from indi.routing import Client, Device

logger = logging.getLogger(__name__)


SenderType = Optional[Union[Client, Device]]


class Router:
    """Message router

    Passess messages between device drivers and client connections.
    """

    _instance = None

    DEFAULT_BLOB_POLICY = const.BLOBEnable.NEVER

    def __init__(self) -> None:
        self.clients: List[Client] = []
        self.devices: List[Device] = []
        self.blob_routing: Dict[
            SenderType, Dict[Optional[str], const.BLOBEnableType]
        ] = {}

    @classmethod
    def instance(cls):
        if not cls._instance:
            cls._instance = cls()
        return cls._instance

    def register_device(self, device: Device):
        self.devices.append(device)

    def register_client(self, client: Client):
        logger.debug("Router: registering client %s", client)
        self.clients.append(client)
        self.blob_routing[client] = {}

    def unregister_client(self, client: Client):
        logger.debug("Router: unregistering client %s", client)
        if client in self.clients:
            self.clients.remove(client)
        if client in self.blob_routing:
            del self.blob_routing[client]

    def process_message(self, message: IndiMessage, sender: SenderType = None):
        is_blob = isinstance(message, NewBLOBVector)

        if message.from_client:
            if isinstance(message, EnableBLOB):
                self.process_enable_blob(message, sender)

            for device in self.devices:
                if not device == sender and device.accepts(message.device):
                    device.message_from_client(message)

        if message.from_device:
            for client in self.clients:
                if not client == sender:
                    device_name = getattr(message, "device")
                    client_blob_policy = self.blob_routing.get(client, {}).get(
                        device_name, self.DEFAULT_BLOB_POLICY
                    )
                    if (
                        is_blob
                        and client_blob_policy
                        in (
                            const.BLOBEnable.ALSO,
                            const.BLOBEnable.ONLY,
                        )
                    ) or (not is_blob and client_blob_policy == const.BLOBEnable.NEVER):
                        client.message_from_device(message)

    def process_enable_blob(self, message: EnableBLOB, sender: SenderType):
        self.blob_routing[sender][message.device] = message.value



from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from indi.message.const import State, SwitchState

if TYPE_CHECKING:
    from indi.client.device import Device
    from indi.client.elements import Element
    from indi.client.vectors import Vector


class BaseEvent:
    def __init__(
        self,
        device: Optional[Device] = None,
        vector: Optional[Vector] = None,
        element: Optional[Element] = None,
    ):
        self.device = device
        self.vector = vector
        self.element = element

    def __str__(self):
        res = f"Event <{self.__class__.__name__}>:"
        props = (
            "device",
            "vector",
            "element",
        )
        for prop in props:
            obj = getattr(self, prop)
            if obj:
                res += f" {prop}={obj.name}"
        return res


class DefinitionUpdate(BaseEvent):
    def __init__(self, vector: Vector):
        super().__init__(device=vector.device, vector=vector)


class ValueUpdate(BaseEvent):
    def __init__(self, element: Element, old_value, new_value):
        super().__init__(
            device=element.vector.device, vector=element.vector, element=element
        )
        self.old_value = old_value
        self.new_value = new_value

    def __str__(self):
        from indi.client import elements

        old_value = str(self.old_value) if self.old_value is not None else "∅"
        new_value = str(self.new_value) if self.new_value is not None else "∅"

        if isinstance(self.element, elements.Switch):
            switch_icons = {
                None: "∅",
                SwitchState.ON: "🌕",
                SwitchState.OFF: "🌑",
            }
            old_value = switch_icons[self.old_value]
            new_value = switch_icons[self.new_value]

        if isinstance(self.element, elements.Light):
            light_icons = {
                None: "∅",
                State.IDLE: "⚪️",
                State.OK: "🟢",
                State.BUSY: "🟡",
                State.ALERT: "🔴",
            }
            old_value = light_icons[self.old_value]
            new_value = light_icons[self.new_value]

        old_value = "\033[1m" + old_value + "\033[0m"
        new_value = "\033[1m" + new_value + "\033[0m"
        return super().__str__() + f" {old_value} → {new_value}"


class StateUpdate(BaseEvent):
    def __init__(self, vector: Vector, old_state, new_state):
        super().__init__(device=vector.device, vector=vector)
        self.old_state = old_state
        self.new_state = new_state

    def __str__(self):
        state_icons = {
            None: "∅",
            State.IDLE: "⚪️",
            State.OK: "🟢",
            State.BUSY: "🟡",
            State.ALERT: "🔴",
        }
        return (
            super().__str__()
            + f" {state_icons[self.old_state]} → {state_icons[self.new_state]}"
        )



from __future__ import annotations

import asyncio
import uuid
from typing import TYPE_CHECKING, Callable, Dict, List, Optional, Type, Union

if TYPE_CHECKING:
    from typing_extensions import Protocol

    from indi.device.driver import Driver
    from indi.device.properties.instance.elements import Element
    from indi.device.properties.instance.vectors import Vector
else:
    Protocol = object

EventCallbackType = Callable[["BaseEvent"], None]


class EventSourceDefinition:
    def __init__(self) -> None:
        self.event_handlers: Dict[
            Type[BaseEvent], Dict[uuid.UUID, EventCallbackType]
        ] = {}
        super().__init__()

    def attach_event_handler(
        self, event_type: Type[BaseEvent], callback: EventCallbackType
    ) -> uuid.UUID:
        if event_type not in self.event_handlers:
            self.event_handlers[event_type] = {}

        uid = uuid.uuid4()
        self.event_handlers[event_type][uid] = callback
        return uid

    def detach_event_handler(self, uid: uuid.UUID):
        for e in self.event_handlers:
            if uid in self.event_handlers[e]:
                del self.event_handlers[e][uid]


class EventSourceProtocol(Protocol):
    @property
    def _definition(self) -> EventSourceDefinition:
        ...


class EventSource:
    """Mixin used in all event sources (Device, Vector, Element))"""

    def raise_event(self: EventSourceProtocol, event: BaseEvent):
        event_type = event.__class__
        callbacks = self._definition.event_handlers.get(event_type, {})

        for uid, cb in callbacks.items():
            if asyncio.iscoroutinefunction(cb):
                asyncio.get_running_loop().create_task(cb(event))
            else:
                cb(event)


def attach_event_handlers(obj: Driver):
    """Attaches event handlers marked by @on(...) decorator.

    Meant to be used in device definition constuctor.
    Attaches event handlers defined as instance method during definition object initialization.
    It is important to call it at the initialization stage, so `self` argument of event handlers is not broken.

    :param obj: Instance of device definition
    :type obj: object
    """
    for f in [
        getattr(obj, fn)
        for fn in dir(obj)
        if hasattr(obj, fn) and callable(getattr(obj, fn))
    ]:
        if hasattr(f, "event_handler_attachments"):
            attachments = f.event_handler_attachments
            assert isinstance(attachments, list)

            for attachment in attachments:
                assert isinstance(attachment, EventHandlerAttachment)
                attachment.src.attach_event_handler(attachment.event_type, f)


class EventHandlerAttachment:
    def __init__(self, src: EventSourceDefinition, event_type: Type[BaseEvent]) -> None:
        self.src = src
        self.event_type = event_type


def on(
    src: Union[EventSourceDefinition, List[EventSourceDefinition]],
    event_type: Type[BaseEvent],
):
    """Decorator with arguments used to mark instance methods as events handlers.

    Marks device definition instance method to bo attached as event handler during device initialization.

    :param src: Event source or list of event sources
    :type src: Union[EventSource, List[EventSource]]
    :param event_type: Event type
    :type event_type: type
    """

    def wrapper(fn: EventCallbackType):
        if not hasattr(fn, "event_handler_attachments"):
            event_handler_attachments: List[EventHandlerAttachment] = []
            setattr(fn, "event_handler_attachments", event_handler_attachments)

        if isinstance(src, list):
            sources = src
        else:
            sources = [src]

        for s in sources:
            getattr(fn, "event_handler_attachments").append(
                EventHandlerAttachment(s, event_type)
            )

        return fn

    return wrapper


class BaseEvent:
    """Base class for all events."""

    def __init__(
        self,
        device: Optional[Driver] = None,
        vector: Optional[Vector] = None,
        element: Optional[Element] = None,
    ) -> None:
        self.device = device
        self.vector = vector
        self.element = element
        self.prevent_default = False
        self.propagate = True


class Write(BaseEvent):
    """Event raised after receiving new value from client.

    Can be used to write new value to physical device.
    """

    def __init__(self, element: Element, new_value) -> None:
        super().__init__(
            element=element, vector=element.vector, device=element.vector.device
        )
        self.new_value = new_value


class Read(BaseEvent):
    """Event raised before sending value to client.

    Can be used to read value from physical device.
    """

    def __init__(self, element: Element) -> None:
        super().__init__(
            element=element, vector=element.vector, device=element.vector.device
        )


class Change(BaseEvent):
    """Event raised on value change."""

    def __init__(self, element: Element, old_value, new_value) -> None:
        super().__init__(
            element=element, vector=element.vector, device=element.vector.device
        )
        self.new_value = new_value
        self.old_value = old_value



import asyncio
import logging
from typing import Callable

from indi.message import IndiMessage
from indi.transport import Buffer

logger = logging.getLogger(__name__)


class ConnectionHandler:
    def __init__(
        self,
        reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter,
        callback: Callable[[IndiMessage], None],
    ):
        self.buffer = Buffer()
        self.reader, self.writer = reader, writer
        self.callback = callback
        self.sender_lock = asyncio.Lock()

    async def wait_for_messages(self):
        while True:
            logger.debug("TCP: waiting for data")
            message = await self.reader.read(1024)
            if not message:
                logger.debug("TCP: no data, breaking")
                break
            logger.debug("TCP: got data: %s", message)
            self.buffer.append(message.decode("latin1"))
            self.buffer.process(self.message_from_server)

    def message_from_server(self, message: IndiMessage):
        self.callback(message)

    def send_message(self, message: IndiMessage):
        data = message.to_string()
        asyncio.get_running_loop().create_task(self.send(data))

    async def send(self, data: bytes):
        async with self.sender_lock:
            logger.debug("TCP: sending data: %s", data)
            self.writer.write(data)
            await self.writer.drain()

    def close(self):
        self.writer.close()


class TCP:
    def __init__(self, address: str = "127.0.0.1", port: int = 7624):
        self.address = address
        self.port = port

    async def connect(self, callback: Callable[[IndiMessage], None]):
        reader, writer = await asyncio.open_connection(self.address, self.port)
        handler = ConnectionHandler(reader, writer, callback)
        return handler



from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from indi.message.base import IndiMessage

if TYPE_CHECKING:
    from indi.message import TimestampType


class DelProperty(IndiMessage):
    from_device = True

    def __init__(
        self,
        device: str,
        name: Optional[str] = None,
        timestamp: Optional[TimestampType] = None,
        message: Optional[str] = None,
        **junk,
    ):
        super().__init__(device)
        self.name = name
        self.timestamp = timestamp
        self.message = message



from typing import Optional

from indi.client.client import BaseClient
from indi.message import IndiMessage
from indi.routing.client import Client as RoutingClient
from indi.routing.router import Router


class SnoopingClient(BaseClient, RoutingClient):
    def __init__(self, router: Optional[Router]) -> None:
        super().__init__()
        self.router = router

    def send_message(self, msg: IndiMessage):
        if msg and self.router:
            self.router.process_message(msg, self)

    def message_from_device(self, message: IndiMessage):
        self.process_message(message)



import datetime
from typing import NewType

from .base import IndiMessage, Message
from .defs import (
    DefBLOBVector,
    DefLightVector,
    DefNumberVector,
    DefSwitchVector,
    DefTextVector,
    DefVector,
)
from .del_property import DelProperty
from .enable_blob import EnableBLOB
from .get_properties import GetProperties
from .news import (
    NewBLOBVector,
    NewNumberVector,
    NewSwitchVector,
    NewTextVector,
    NewVector,
)
from .one_light import OneLight
from .pings import PingReply, PingRequest
from .sets import (
    SetBLOBVector,
    SetLightVector,
    SetNumberVector,
    SetSwitchVector,
    SetTextVector,
    SetVector,
)

TimestampType = NewType("TimestampType", str)


def now() -> TimestampType:
    return TimestampType(datetime.datetime.utcnow().isoformat())



from indi.device import properties
from indi.device.properties import const


def AbsolutePosition(
    label=None,
    default=0,
    min=0,
    max=1000,
    step=1,
    state=const.State.OK,
    perm=const.Permissions.READ_WRITE,
    timeout=0,
    enabled=True,
):
    return properties.NumberVector(
        name="ABS_FOCUS_POSITION",
        elements=dict(
            position=properties.Number(
                name="FOCUS_ABSOLUTE_POSITION",
                default=default,
                min=min,
                max=max,
                step=step,
            )
        ),
    )


def RelativePosition(
    label=None,
    min=1,
    max=1000,
    step=1,
    state=const.State.OK,
    perm=const.Permissions.READ_WRITE,
    timeout=0,
    enabled=True,
):
    return properties.NumberVector(
        name="REL_FOCUS_POSITION",
        elements=dict(
            position=properties.Number(
                name="FOCUS_RELATIVE_POSITION", min=min, max=max, step=step
            )
        ),
    )


def FocusMax(
    label=None,
    default=1000,
    state=const.State.OK,
    perm=const.Permissions.READ_ONLY,
    timeout=0,
    enabled=True,
):
    return properties.NumberVector(
        name="FOCUS_MAX",
        elements=dict(max=properties.Number(name="FOCUS_MAX_VALUE", default=default)),
    )


def FocusMotion(
    label=None,
    default="FOCUS_INWARD",
    state=const.State.OK,
    perm=const.Permissions.READ_WRITE,
    timeout=0,
    enabled=True,
):
    return properties.SwitchVector(
        name="FOCUS_MOTION",
        default_on=default,
        rule=const.SwitchRule.ONE_OF_MANY,
        elements=dict(
            inward=properties.Switch(name="FOCUS_INWARD"),
            outward=properties.Switch(name="FOCUS_OUTWARD"),
        ),
    )



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Dict, Optional, Type, cast

from indi import message
from indi.device.events import attach_event_handlers
from indi.device.properties import Group as GroupDefinition
from indi.device.properties.instance.group import Group, GroupGetter
from indi.device.properties.instance.vectors import Vector
from indi.message import IndiMessage
from indi.routing import Device, Router

if TYPE_CHECKING:
    from indi.device.snoop import SnoopingClient

logger = logging.getLogger(__name__)


class DriverMeta(type):
    def __new__(meta, name, bases, dct):
        groups = {}
        for k in dct:
            if isinstance(dct[k], GroupDefinition):
                groups[k] = dct[k]
                dct[k] = GroupGetter(k)
        dct["_group_definitions"] = groups

        return super().__new__(meta, name, bases, dct)


class Driver(Device, metaclass=DriverMeta):
    _group_definitions: Dict[str, GroupDefinition] = {}

    def __init__(self, name: Optional[str] = None, router: Optional[Router] = None):
        attach_event_handlers(self)

        self._name: str = name or str(self.__class__.name)

        self._groups: Dict[str, Group] = {
            k: Group(self, v)
            for k, v in self.__class__._all_group_definitions().items()
        }
        self._router = router

        self._vectors: Dict[str, Vector] = {}
        for k, group in self._groups.items():
            for k, vector in group.vectors.items():
                self._vectors[vector.name] = vector

        if self._router:
            self._router.register_device(self)
        self._snooping_client: Optional[SnoopingClient] = None

    @property
    def snooping_client(self) -> SnoopingClient:
        from indi.device.snoop import SnoopingClient

        assert (
            self._router
        ), "Router has to be defined before creating a snooping client"

        if self._snooping_client is None:
            self._snooping_client = SnoopingClient(self._router)
            self._router.register_client(self._snooping_client)
        return self._snooping_client

    def snoop_device(self, device: str, name: Optional[str] = None) -> SnoopingClient:
        client = self.snooping_client
        client.handshake(device, name)
        return client

    @classmethod
    def _all_group_definitions(cls) -> Dict[str, GroupDefinition]:
        groups: Dict[str, GroupDefinition] = {}
        for base in cls.__bases__:
            if issubclass(base, Driver) or base is Driver:
                groups = {**groups, **cast(Type[Driver], base)._group_definitions}
        for k, v in cls._group_definitions.items():
            if isinstance(v, GroupDefinition):
                groups[k] = v
        return groups

    @property
    def name(self) -> str:
        return self._name

    def accepts(self, device: Optional[str]) -> bool:
        return device is None or self.name == device

    def get_group(self, name: str) -> Optional[Group]:
        return self._groups.get(name)

    def send_message(self, msg: Optional[IndiMessage]):
        if self._router and msg:
            self._router.process_message(msg, self)

    def message_from_client(self, msg: IndiMessage):
        if isinstance(msg, message.GetProperties):
            if not msg.name:
                for k, v in self._vectors.items():
                    self.send_message(v.to_def_message())
            else:
                if msg.name in self._vectors:
                    v = self._vectors[msg.name]
                    self.send_message(v.to_def_message())

        if isinstance(msg, message.news.NewVector):
            self._vectors[msg.name].from_new_message(msg)



from typing import Dict, Optional

from indi.device.events import EventSourceDefinition
from indi.device.properties.definition.vectors import Vector


class Group(EventSourceDefinition):
    def __init__(
        self, name: str, enabled=True, vectors: Optional[Dict[str, Vector]] = None
    ):
        super().__init__()
        self.name = name
        self.vectors: Dict[str, Vector] = vectors or {}
        self.enabled = enabled

    def __getattr__(self, item: str):
        vector = self.vectors.get(item)
        if vector:
            return vector

        return self.__getattribute__(item)



from __future__ import annotations

from typing import TYPE_CHECKING, Dict

from indi.device.properties.definition.group import Group as GroupDefinition
from indi.device.properties.instance.vectors import Vector

if TYPE_CHECKING:
    from indi.device.driver import Driver


class Group:
    def __init__(self, device: Driver, definition: GroupDefinition) -> None:
        self._device = device
        self._definition: GroupDefinition = definition
        self._vectors: Dict[str, Vector] = {
            k: v.instance(self) for k, v in definition.vectors.items()
        }
        self._enabled = definition.enabled

    def __getattr__(self, item):
        vector = self._vectors.get(item)
        if vector:
            return vector

        return self.__getattribute__(item)

    @property
    def device(self) -> Driver:
        return self._device

    @property
    def name(self) -> str:
        return self._definition.name

    @property
    def vectors(self) -> Dict[str, Vector]:
        return self._vectors

    @property
    def enabled(self) -> bool:
        return self._enabled

    @enabled.setter
    def enabled(self, value: bool):
        self._enabled = value
        for k, v in self._vectors.items():
            self.device.send_message(v.to_def_message())
            self.device.send_message(v.to_set_message())


class GroupGetter:
    def __init__(self, property_name: str):
        self.property_name = property_name

    def __get__(self, instance: Driver, objtype=None):
        assert self.property_name
        return instance.get_group(self.property_name)



from indi.message.base import IndiMessage


class PingReply(IndiMessage):
    from_client = True

    def __init__(self, uid: str, **junk):
        super().__init__()
        self.uid = uid


class PingRequest(IndiMessage):
    from_device = True

    def __init__(self, uid: str, **junk):
        super().__init__()
        self.uid = uid



from __future__ import annotations

import xml.etree.cElementTree as ET
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Type

if TYPE_CHECKING:
    from indi.message import TimestampType


class IndiMessage:
    from_device = False
    from_client = False

    def __init__(self, device=None, **junk):
        self.device = device

    @classmethod
    def tag_name(cls):
        return cls.__name__[:1].lower() + cls.__name__[1:]

    @classmethod
    def __all_subclasses__(cls) -> Tuple[Type[IndiMessage], ...]:
        subclasses = []
        for subclass in cls.__subclasses__():
            subclasses.append(subclass)
            for nested_subclass in subclass.__all_subclasses__():
                subclasses.append(nested_subclass)
        return tuple(subclasses)

    @classmethod
    def from_xml(cls, xml: ET.Element) -> IndiMessage:
        tag = xml.tag
        message_class = None

        for subclass in cls.__all_subclasses__():
            if subclass.tag_name() == tag:
                message_class = subclass

        if not message_class:
            raise Exception(f"Invalid message: {tag}")

        kwargs: Dict[str, Any] = xml.attrib

        children: Tuple[ET.Element, ...] = tuple(
            IndiMessagePart.from_xml(child) for child in xml
        )
        if children:
            kwargs["children"] = children

        if xml.text:
            kwargs["value"] = xml.text.strip()

        return message_class(**kwargs)

    @classmethod
    def from_string(cls, string: str) -> IndiMessage:
        xml = ET.fromstring(string)
        return cls.from_xml(xml)

    def to_xml(self):
        kwargs = {
            k: str(v)
            for k, v in sorted(self.__dict__.items())
            if v is not None
            and k
            not in (
                "children",
                "value",
            )
        }

        element = ET.Element(self.__class__.tag_name(), **kwargs)

        if getattr(self, "value", None) is not None:
            element.text = str(self.value)

        if hasattr(self, "children"):
            for child in self.children:
                child.to_xml(element)

        return element

    def to_string(self) -> bytes:
        xml = self.to_xml()
        return b'<?xml version="1.0"?>\n' + ET.tostring(xml) + b"\n"

    def to_dict(self):
        res = {
            k: str(v)
            for k, v in sorted(self.__dict__.items())
            if v is not None
            and k
            not in (
                "children",
                "value",
            )
        }

        if getattr(self, "value", None) is not None:
            res["_value"] = str(self.value)

        if hasattr(self, "children"):
            for child in self.children:
                res["_children"] = child.to_dict()

        return res

    def __eq__(self, other):
        return other.__class__ == self.__class__ and self.to_dict() == other.to_dict()


class Message(IndiMessage):
    from_device = True

    def __init__(
        self,
        device: Optional[str] = None,
        timestamp: Optional[TimestampType] = None,
        message: Optional[str] = None,
        **junk,
    ):
        super().__init__(device)
        self.timestamp = timestamp
        self.message = message


class IndiMessagePart:
    def __init__(self, name, value, **junk):
        self.name = name
        self.value = self.check_value(value)

    def check_value(self, value):
        return value

    @classmethod
    def tag_name(cls):
        return cls.__name__[:1].lower() + cls.__name__[1:]

    @classmethod
    def _all_subclasses(cls):
        all_subclasses = []

        for subclass in cls.__subclasses__():
            all_subclasses.append(subclass)
            all_subclasses.extend(subclass._all_subclasses())

        return set(all_subclasses)

    @classmethod
    def from_xml(cls, xml):
        tag = xml.tag
        message_class = None

        for subclass in cls._all_subclasses():
            if subclass.tag_name() == tag:
                message_class = subclass

        if not message_class:
            raise Exception(f"Invalid part: {tag}")

        kwargs = xml.attrib

        kwargs["value"] = xml.text.strip() if xml.text else None

        return message_class(**kwargs)

    def to_xml(self, parent):
        kwargs = {
            k: str(v)
            for k, v in self.__dict__.items()
            if v is not None and k not in ("value",)
        }

        element = ET.SubElement(parent, self.__class__.tag_name(), **kwargs)
        if self.value is not None:
            element.text = str(self.value)

        return element

    def to_dict(self):
        res = {
            k: str(v)
            for k, v in sorted(self.__dict__.items())
            if v is not None and k not in ("value",)
        }

        if getattr(self, "value", None) is not None:
            res["_value"] = str(self.value)

        return res

    def __eq__(self, other):
        return other.__class__ == self.__class__ and self.to_dict() == other.to_dict()



from typing import Dict, Optional, Tuple, Type, Union

from indi.device.events import EventSourceDefinition
from indi.device.properties.const import (
    Permissions,
    PermissionsType,
    State,
    StateType,
    SwitchRule,
    SwitchRuleType,
)
from indi.device.properties.definition.elements import (
    BLOB,
    Element,
    Light,
    Number,
    Switch,
    Text,
)
from indi.device.properties.instance import vectors as instance_vectors
from indi.message import const


class Vector(EventSourceDefinition):
    element_class: Type[Element]
    instance_class: Type[instance_vectors.Vector]

    def __init__(
        self,
        name: str,
        label: Optional[str] = None,
        state: StateType = State.OK,
        perm: PermissionsType = Permissions.READ_WRITE,
        timeout: float = 0,
        enabled: bool = True,
        elements: Optional[Dict[str, Element]] = None,
    ):
        super().__init__()
        self.name = name
        self.label = label or name
        self.state = state
        self.perm = perm
        self.timeout = timeout
        self.enabled = enabled

        if not elements:
            raise Exception("No vector elements declared")

        for k, element in elements.items():
            if self.__class__.element_class and not isinstance(
                element, self.__class__.element_class
            ):
                raise Exception(
                    f"Element '{k}' of vector {self.name} of type {self.__class__.__name__} "
                    f"is expected to be of type {self.__class__.element_class.__name__}, "
                    f"{element.__class__.__name__} given"
                )

        self.elements = elements

    def instance(self, group):
        return self.instance_class(group=group, definition=self)

    def __getattr__(self, item):
        element = self.elements.get(item)
        if element:
            return element

        return self.__getattribute__(item)


class NumberVector(Vector):
    element_class = Number
    instance_class = instance_vectors.NumberVector


class TextVector(Vector):
    element_class = Text
    instance_class = instance_vectors.TextVector


class SwitchVector(Vector):
    element_class = Switch
    instance_class = instance_vectors.SwitchVector

    RULES = SwitchRule

    def __init__(
        self,
        *args,
        rule: SwitchRuleType = SwitchRule.ONE_OF_MANY,
        default_on: Optional[Union[str, Tuple[str, ...]]] = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.rule = rule
        if default_on is not None:
            default_on = (default_on,) if isinstance(default_on, str) else default_on
            for k, el in self.elements.items():
                if el.name in default_on:
                    el.default = const.SwitchState.ON


class LightVector(Vector):
    element_class = Light
    instance_class = instance_vectors.LightVector

    def __init__(self, *args, **kwargs):
        if kwargs.get("perm"):
            del kwargs["perm"]
        if kwargs.get("timeout"):
            del kwargs["timeout"]

        super().__init__(*args, **kwargs)


class BLOBVector(Vector):
    element_class = BLOB
    instance_class = instance_vectors.BLOBVector



from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Tuple, Type, Union

from indi.message import checks
from indi.message.base import IndiMessage
from indi.message.one_parts import (
    IndiMessagePart,
    OneBLOB,
    OneNumber,
    OneSwitch,
    OneText,
)

if TYPE_CHECKING:
    from indi.message import TimestampType


class NewVector(IndiMessage):
    children_class: Union[
        Type[OneBLOB], Type[OneNumber], Type[OneSwitch], Type[OneText]
    ]
    from_client = True

    def __init__(
        self,
        device: str,
        name: str,
        timestamp: Optional[TimestampType] = None,
        children: Optional[Tuple[IndiMessagePart, ...]] = None,
        **junk,
    ):
        super().__init__(device)
        self.name = name
        self.timestamp = timestamp
        self.children = checks.children(children, self.children_class)


class NewBLOBVector(NewVector):
    children_class = OneBLOB


class NewNumberVector(NewVector):
    children_class = OneNumber


class NewSwitchVector(NewVector):
    children_class = OneSwitch


class NewTextVector(NewVector):
    children_class = OneText



from __future__ import annotations

import base64
import hashlib
import math
import re
from typing import Any, Optional


class BLOB:
    def __init__(self, binary: bytes, format: str):
        self.binary = binary
        self.format = format

    @property
    def size(self) -> int:
        return len(self.binary)

    @property
    def binary_base64(self) -> str:
        return base64.b64encode(self.binary).decode("latin1")

    @property
    def md5(self) -> str:
        return hashlib.md5(self.binary).hexdigest()

    @classmethod
    def from_base64(cls, binary_base64: str, format: str) -> BLOB:
        return cls(base64.b64decode(binary_base64), format)

    def __len__(self):
        return self.size


def str_to_num(s: str, fmt: str) -> Any[float, int]:
    if s is None:
        return None
    if not isinstance(s, str):
        s = str(s)

    sexagesimal_match = re.match(r"^%(\d*)\.(\d+)m$", fmt)
    if sexagesimal_match:
        fraction_length = int(sexagesimal_match.groups()[1])
        assert fraction_length in (
            3,
            5,
            6,
            8,
            9,
        ), f"Invalid sexagesimal number format: {fmt}"

        regexps = {
            3: r"^(\-?\d+)[:; ](\d{2})$",
            5: r"^(\-?\d+)[:; ](\d{2}\.\d+)$",
            6: r"^(\-?\d+)[:; ](\d{2})[:; ](\d{2})$",
            8: r"^(\-?\d+)[:; ](\d{2})[:; ](\d{2}.\d+)$",
            9: r"^(\-?\d+)[:; ](\d{2})[:; ](\d{2}.\d+)$",
        }

        num_match = re.match(regexps[fraction_length], s)
        if not num_match:
            raise ValueError("Cannot convert string to number")
        num_match_groups = num_match.groups()
        wholes = num_match_groups[0]
        minutes = num_match_groups[1]
        seconds = num_match_groups[2] if fraction_length in (6, 8, 9) else 0

        return float(wholes) + (float(minutes) / 60) + (float(seconds) / 3600)

    if "." in s:
        return float(s)

    return int(s)


def num_to_str(n: Optional[float], fmt: str) -> Optional[str]:
    if n is None:
        return None

    sexagesimal_match = re.match(r"^%(\d*)\.(\d+)m$", fmt)
    if sexagesimal_match:
        fraction_length = int(sexagesimal_match.groups()[1])
        assert fraction_length in (3, 5, 6, 8, 9)

        w = math.floor(n)
        m = (n - w) * 60

        if fraction_length == 3:
            return f"{w}:{m:02.0f}"

        if fraction_length == 5:
            return f"{w}:{m:04.1f}"

        mf = math.floor(m)
        s = (m - mf) * 60
        m = mf

        if fraction_length == 6:
            return f"{w}:{m:02d}:{s:02.0f}"

        if fraction_length == 8:
            return f"{w}:{m:02d}:{s:04.1f}"

        if fraction_length == 9:
            return f"{w}:{m:02d}:{s:05.2f}"

    return fmt % n



from __future__ import annotations

import asyncio
import logging
import uuid
from typing import Any, Callable, Dict, Iterable, List, Optional, Type

import indi
from indi import message
from indi.client import events
from indi.client.device import Device
from indi.message import IndiMessage, const

logger = logging.getLogger(__name__)


class _CallbackConfig:
    def __init__(
        self,
        device: Optional[str],
        vector: Optional[str],
        element: Optional[str],
        event_type: Type[events.BaseEvent],
        callback: Callable,
        uuid: uuid.UUID,
    ):
        self.device = device
        self.vector = vector
        self.element = element
        self.event_type = event_type
        self.callback = callback
        self.uuid = uuid

    def accepts_event(self, event: events.BaseEvent) -> bool:
        """Checks if event should be processed by callbacked associated with this configuration.

        :param event: An event
        :type event: events.BaseEvent
        :return: True if event should be processed
        :rtype: bool
        """
        return (
            self.device
            in (
                None,
                event.device.name if event.device else None,
            )
            and self.vector
            in (
                None,
                event.vector.name if event.vector else None,
            )
            and self.element
            in (
                None,
                event.element.name if event.element else None,
            )
            and isinstance(event, self.event_type)
        )


class _EventWaitResult:
    def __init__(self) -> None:
        self.event: Optional[events.BaseEvent] = None
        self.timeout = False


class BaseClient:
    def __init__(self) -> None:
        """Constructor for INDI client."""
        self.devices: Dict[str, Device] = {}
        self.callbacks: List[_CallbackConfig] = []

    def __getitem__(self, key) -> Device:
        return self.devices[key]

    def __contains__(self, key) -> bool:
        return key in self.devices

    def get_device(self, name: str) -> Optional[Device]:
        return self.devices.get(name)

    def list_devices(self) -> Iterable[str]:
        """Lists all known device names.

        :return: List of all known device names
        :rtype: Iterable[str]
        """
        return self.devices.keys()

    def set_device(self, name: str, device: Device):
        self.devices[name] = device
        self.blob_handshake(name)

    def process_message(self, msg: IndiMessage):
        device = None
        if isinstance(msg, message.DefVector):
            device = self.get_device(msg.device)
            if not device:
                device = Device(self, msg.device)
                self.set_device(msg.device, device)

        if isinstance(msg, message.SetVector):
            device = self.get_device(msg.device)

        if isinstance(msg, message.DelProperty):
            device = self.get_device(msg.device)

        if device:
            device.process_message(msg)

    def send_message(self, msg: IndiMessage):
        """Sends INDI message to server using control connection.

        :param msg: INDI message to be sent
        :type msg: IndiMessage
        """
        raise NotImplementedError()

    def onevent(
        self,
        *,
        callback: Callable,
        device: Optional[str] = None,
        vector: Optional[str] = None,
        element: Optional[str] = None,
        event_type: Type[events.BaseEvent] = events.BaseEvent,
    ) -> uuid.UUID:
        """Attaches event callback.

        :param callback: Callback
        :type callback: Callable
        :param device: Optional device name, defaults to None
        :type device: str, optional
        :param vector: Optional vector name, defaults to None
        :type vector: str, optional
        :param element: Optional element name, defaults to None
        :type element: str, optional
        :param event_type: Optional event type, defaults to events.BaseEvent
        :type event_type: Type[events.BaseEvent], optional
        :return: UUID of created event attachment
        :rtype: uuid
        """
        uid = uuid.uuid4()
        callback_config = _CallbackConfig(
            device=device,
            vector=vector,
            element=element,
            event_type=event_type,
            callback=callback,
            uuid=uid,
        )

        self.callbacks.append(callback_config)
        return uid

    def rmonevent(
        self,
        uuid: Optional[uuid.UUID] = None,
        device: Optional[str] = None,
        vector: Optional[str] = None,
        element: Optional[str] = None,
        event_type: Optional[Type[events.BaseEvent]] = None,
        callback: Optional[Callable] = None,
    ):
        to_rm = list()
        for cb in self.callbacks:
            if (
                uuid
                in (
                    None,
                    cb.uuid,
                )
                and device
                in (
                    None,
                    cb.device,
                )
                and vector
                in (
                    None,
                    cb.vector,
                )
                and element
                in (
                    None,
                    cb.element,
                )
                and event_type
                in (
                    None,
                    cb.event_type,
                )
                and callback in (None, cb.callback)
            ):
                to_rm.append(cb)

        for cb in to_rm:
            self.callbacks.remove(cb)

    async def waitforevent(
        self,
        device: Optional[str] = None,
        vector: Optional[str] = None,
        element: Optional[str] = None,
        event_type: Type[events.BaseEvent] = events.BaseEvent,
        expect: Any = None,
        initial: Any = None,
        check: Optional[Callable] = None,
        timeout: Optional[Optional[float]] = None,
        polling_enabled: bool = True,
        polling_delay: float = 1.0,
        polling_interval: float = 1.0,
    ):
        assert 1 == sum(
            1
            for _ in filter(
                None.__ne__,
                (
                    expect,
                    initial,
                    check,
                ),
            )
        ), "Exactly one of `expect`, `initial`, `check` has to be passed"

        lock = asyncio.Event()
        result = _EventWaitResult()

        def cb(event: events.BaseEvent):
            release = False

            if check is not None:
                if check(event):
                    release = True

            if expect is not None:
                if isinstance(event, events.ValueUpdate):
                    if event.new_value == expect:
                        release = True
                if isinstance(event, events.StateUpdate):
                    if event.new_state == expect:
                        release = True

            if initial is not None:
                if isinstance(event, events.ValueUpdate):
                    if event.new_value != initial:
                        release = True
                if isinstance(event, events.StateUpdate):
                    if event.new_state != initial:
                        release = True

            if release:
                result.event = event
                lock.set()

        if polling_enabled:

            async def poll():
                kwargs = {}
                if device:
                    kwargs["device"] = device
                if vector:
                    kwargs["name"] = vector

                msg = message.GetProperties(version=indi.__protocol_version__, **kwargs)

                await asyncio.sleep(polling_delay)
                while not lock.is_set():
                    self.send_message(msg)
                    await asyncio.sleep(polling_interval)

            asyncio.get_running_loop().create_task(poll())

        async def timeout_check():
            await asyncio.sleep(timeout)
            if not lock.is_set():
                result.timeout = True
                lock.set()

        uid = self.onevent(
            device=device,
            vector=vector,
            element=element,
            event_type=event_type,
            callback=cb,
        )

        if timeout is not None and timeout > 0:
            asyncio.get_running_loop().create_task(timeout_check())

        await lock.wait()

        self.rmonevent(uuid=uid)

        if result.timeout:
            raise Exception("Timeout occurred")

        return result.event

    def trigger_event(self, event: events.BaseEvent):
        for callback in self.callbacks:
            if callback.accepts_event(event):
                try:
                    callback.callback(event)
                except:
                    logger.exception("Error in event handler")

    def handshake(self, device=None, name=None, version=indi.__protocol_version__):
        self.send_message(
            message.GetProperties(version=version, device=device, name=name)
        )

    def blob_handshake(self, device):
        self.send_message(
            message.EnableBLOB(device=device, value=const.BLOBEnable.NEVER)
        )


class Client(BaseClient):
    def __init__(self, control_connection, blob_connection):
        super().__init__()
        self.control_connection = control_connection
        self.blob_connection = blob_connection
        self.control_connection_handler = None
        self.blob_connection_handler = None

    def send_message(self, msg: IndiMessage):
        """Sends INDI message to server using control connection.

        :param msg: INDI message to be sent
        :type msg: IndiMessage
        """
        self.control_connection_handler.send_message(msg)

    def blob_handshake(self, device):
        super().blob_handshake(device)
        self.blob_connection_handler.send_message(
            message.EnableBLOB(device=device, value=const.BLOBEnable.ONLY)
        )

    async def start(self):
        """Starts client and connects to the server.

        Connects both connections (control and blob) and sends initial GetProperties message to the server.
        """
        self.control_connection_handler = await self.control_connection.connect(
            self.process_message
        )
        self.blob_connection_handler = await self.blob_connection.connect(
            self.process_message
        )

        asyncio.get_running_loop().create_task(
            self.control_connection_handler.wait_for_messages()
        )
        asyncio.get_running_loop().create_task(
            self.blob_connection_handler.wait_for_messages()
        )

        self.handshake()

    def stop(self):
        self.control_connection_handler.close()
        self.blob_connection_handler.close()



import re
from pathlib import Path
from os import path

from setuptools import find_packages, setup

BUNDLES = ()

# -*- Installation Requires -*-


def strip_comments(l):
    return l.split("#", 1)[0].strip()


def _pip_requirement(req, *root):
    if req.startswith("-r "):
        _, path = req.split()
        return reqs(*root, *path.split("/"))
    return [req]


def _reqs(*f):
    path = (Path.cwd() / "requirements").joinpath(*f)
    with path.open() as fh:
        reqs = [strip_comments(l) for l in fh.readlines()]
        return [_pip_requirement(r, *f[:-1]) for r in reqs if r]


def reqs(*f):
    return [req for subreq in _reqs(*f) for req in subreq]


def extras(*p):
    """Parse requirement in the requirements/extras/ directory."""
    return reqs("extras", *p)


def extras_require():
    """Get map of all extra requirements."""
    return {x: extras(x + ".txt") for x in BUNDLES}


def long_description():
    with open("README.md", "r") as fh:
        return fh.read()


def get_about():
    """Parses __init__ on main module in search of all dunder names"""
    regex = re.compile(r"^__\w+__\s*=.*$")
    about = dict()
    with open("indi/__init__.py", "r") as f:
        dunders = list()
        for l in f.readlines():
            if regex.match(l):
                dunders.append(l)
        exec("\n".join(dunders), about)

    with open(path.join(path.dirname(__file__), "indi", "VERSION")) as f:
        about["__version__"] = f.read().strip()

    return about

about = get_about()


setup(
    name="INDIpy",
    version=about["__version__"],
    description="Python implementation of INDI server and client",
    url="http://github.com/wlatanowicz/indipy",
    author=about["__author__"],
    author_email="indipy@wiktor.latanowicz.com",
    license="MIT",
    long_description=long_description(),
    long_description_content_type="text/markdown",
    packages=find_packages(exclude=["tests*", "docker-examples"]),
    zip_safe=False,
    install_requires=reqs("base.txt"),
    tests_require=reqs("tests.txt"),
    extras_require=extras_require(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8",
    test_suite="tests",
    include_package_data=True,
)



import logging
import xml.etree.ElementTree as ET
from typing import Callable

from indi.message import IndiMessage

logger = logging.getLogger(__name__)


class Buffer:
    def __init__(self) -> None:
        self.data = ""
        self.allowed_tags = [m.tag_name() for m in IndiMessage.__all_subclasses__()]

    def append(self, data: str):
        self.data += data

    def _cleanup_buffer(self):
        start = len(self.data) - 1
        for tag in self.allowed_tags:
            start = min(start, self.data.find("<" + tag))

        if start >= 0:
            self.data = self.data[start:]

    def process(self, callback: Callable[[IndiMessage], None]):
        self._cleanup_buffer()
        end = 0
        while len(self.data) > 0 and end >= 0:
            end = self.data.find(">", end)

            if end > 0:
                end += 1
                partial = self.data[0:end]

                try:
                    ET.fromstring(partial)
                    is_correct_xml = True
                except ET.ParseError:
                    is_correct_xml = False

                if is_correct_xml:
                    self.data = self.data[end:]
                    end = 0
                    message = None
                    try:
                        message = IndiMessage.from_string(partial)
                    except Exception:
                        logger.warning("Buffer: Contents is not a valid message")

                    if message:
                        try:
                            callback(message)
                        except Exception:
                            logger.exception("Error procesing message")



import asyncio
import logging
from typing import List

from indi.message import IndiMessage
from indi.routing import Client, Router
from indi.transport import Buffer

logger = logging.getLogger(__name__)


class ConnectionHandler(Client):
    connections: List["ConnectionHandler"] = []

    def __init__(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter, router: Router
    ):
        self.buffer = Buffer()
        self.reader, self.writer = reader, writer
        self.router = router
        self.sender_lock = asyncio.Lock()
        if self.router:
            self.router.register_client(self)

    @classmethod
    def handler(cls, router: Router):
        async def handler_func(reader, writer):
            conn = cls(reader, writer, router)
            cls.connections.append(conn)
            try:
                await conn.wait_for_messages()
            except:
                logger.exception("Error in client handler loop")

            conn.close()
            cls.connections.remove(conn)

        return handler_func

    async def wait_for_messages(self):
        while True:
            logger.debug(f"TCP: waiting for data")
            message = await self.reader.read(1024)
            if not message:
                logger.debug(f"TCP: no data, breaking")
                break
            logger.debug("TCP: got data: %s", message)
            self.buffer.append(message.decode("latin1"))
            self.buffer.process(self.message_from_client)

    def message_from_client(self, message: IndiMessage):
        if self.router:
            self.router.process_message(message, sender=self)

    def message_from_device(self, message: IndiMessage):
        data = message.to_string()
        asyncio.get_running_loop().create_task(self.send(data))

    async def send(self, data: bytes):
        async with self.sender_lock:
            logger.debug("TCP: sending data: %s", data)
            self.writer.write(data)
            await self.writer.drain()

    def close(self):
        self.writer.close()
        if self.router:
            self.router.unregister_client(self)


class TCP:
    def __init__(self, router: Router, address="0.0.0.0", port=7624):
        self.address = address
        self.port = port
        self.router = router

    async def client_connected(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
    ):
        handler = ConnectionHandler.handler(self.router)
        await handler(reader, writer)

    async def start(self):
        logger.info("Starting async TCP INDI server on %s:%s", self.address, self.port)
        server = await asyncio.start_server(
            self.client_connected, self.address, self.port
        )

        try:
            async with server:
                await server.serve_forever()
        finally:
            for conn in ConnectionHandler.connections:
                conn.close()



from __future__ import annotations

from typing import TYPE_CHECKING, Dict, Optional, Tuple

from indi import message
from indi.client.events import DefinitionUpdate
from indi.client.vectors import Vector

if TYPE_CHECKING:
    from indi.client.client import BaseClient


class Device:
    def __init__(self, client: BaseClient, name: str):
        self.vectors: Dict[str, Vector] = {}
        self.client = client
        self.name = name

    def __getitem__(self, key) -> Vector:
        return self.vectors[key]

    def __contains__(self, key) -> bool:
        return key in self.vectors

    def get_vector(self, name) -> Optional[Vector]:
        return self.vectors.get(name)

    def list_vectors(self) -> Tuple[str, ...]:
        return tuple(str(key) for key in self.vectors.keys())

    def set_vector(self, name: str, vector: Vector):
        self.vectors[name] = vector

    def process_message(self, msg: message.IndiMessage):
        vector = None
        if isinstance(msg, message.DefVector):
            vector = Vector.from_message(self, msg)
            self.set_vector(msg.name, vector)
            event = DefinitionUpdate(vector)
            self.client.trigger_event(event)

        if isinstance(msg, message.SetVector):
            vector = self.get_vector(msg.name)

        if isinstance(msg, message.DelProperty):
            if msg.name in self.vectors:
                del self.vectors[msg.name]
            vector = None

        if vector:
            vector.process_message(msg)

    def send_message(self, msg: message.IndiMessage):
        self.client.send_message(msg)



from typing import Optional, Type

from indi.client.events import ValueUpdate
from indi.device import values
from indi.message import def_parts, one_parts


class Element:
    def_message_class: Optional[Type[def_parts.DefIndiMessagePart]] = None
    set_message_class: Optional[Type[one_parts.IndiMessagePart]] = None
    new_message_class: Optional[Type[one_parts.IndiMessagePart]] = None

    @classmethod
    def from_message(cls, vector, msg):
        for subclass in cls.__subclasses__():
            if isinstance(msg, subclass.def_message_class):
                element = subclass(vector, msg)
                event = ValueUpdate(element, None, element._value)
                vector.device.client.trigger_event(event)
                return element
        return None

    def __init__(self, vector, msg):
        self.name = msg.name
        self.label = msg.label
        self.vector = vector
        self._value = msg.value
        self._new_value = None

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._new_value = value

    @property
    def has_new_value(self) -> bool:
        return self._new_value is not None

    def reset_new_value(self):
        self._new_value = None

    def process_message(self, msg):
        if isinstance(msg, self.set_message_class):
            old_value = self.value
            self.set_value_from_message(msg)
            if self._value != old_value:
                event = ValueUpdate(self, old_value, self._value)
                self.vector.device.client.trigger_event(event)

    def set_value_from_message(self, msg):
        self._value = msg.value

    def to_new_message(self):
        return self.new_message_class(name=self.name, value=self._new_value)


class Number(Element):
    def_message_class = def_parts.DefNumber
    set_message_class = one_parts.OneNumber
    new_message_class = one_parts.OneNumber


class Switch(Element):
    def_message_class = def_parts.DefSwitch
    set_message_class = one_parts.OneSwitch
    new_message_class = one_parts.OneSwitch


class Text(Element):
    def_message_class = def_parts.DefText
    set_message_class = one_parts.OneText
    new_message_class = one_parts.OneText


class Light(Element):
    def_message_class = def_parts.DefLight
    set_message_class = one_parts.OneLight
    new_message_class = one_parts.OneLight


class BLOB(Element):
    def_message_class = def_parts.DefBLOB
    set_message_class = one_parts.OneBLOB
    new_message_class = one_parts.OneBLOB

    def set_value_from_message(self, msg):
        blob_value = values.BLOB.from_base64(msg.value, msg.format)
        assert (
            int(msg.size) == blob_value.size
        ), f"Blob size differs: {msg.size} declared vs {blob_value.size} measured"

        self._value = blob_value



import re


def children(value, child_class):
    if value is None:
        return []
    for ch in value:
        if not isinstance(ch, child_class):
            raise ValueError(
                "Child node has to be of type %s, %s given",
                child_class.__name__,
                ch.__class__.__name__,
            )
    return value


def dictionary(value, dictionary_class):
    if value not in dictionary_class.__dict__.values():
        raise ValueError(
            'Invalid value: "%s" not found in %s', value, dictionary_class.__name__
        )
    return value


def number(value):
    regexps = (
        r"^\-?\d+$",  # int
        r"^\-?\d+\.\d+$",
        r"^\-?\d+\.$",
        r"^\-?\.\d+$",  # float
        r"^\-?\d+:\d{2}$",  # :mm
        r"^\-?\d+:\d{2}\.\d+$",  # :mm.m
        r"^\-?\d+:\d{2}:\d{2}$",  # :mm:ss
        r"^\-?\d+:\d{2}:\d{2}\.\d+$",  # :mm:ss.s
    )
    if value is None:
        return None

    if not any([re.match(r, str(value)) for r in regexps]):
        raise ValueError("Invalid value for number: %s", value)

    return value
